package com.ten60.netkernel.cache.se.resolution;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.atomic.AtomicBoolean;


import java.io.*;

import org.netkernel.container.IKernel;
import org.netkernel.container.IResolutionCache;
import org.netkernel.container.config.IConfiguration;
import org.netkernel.container.config.IConfigurationListener;
import org.netkernel.request.IRequest;
import org.netkernel.request.impl.RequestFactory;
import org.netkernel.urii.IResolution;
import org.netkernel.util.Utils;
import org.netkernel.util.XMLUtils;

import com.ten60.netkernel.cache.se.representation2.ConcurrentCache;

public class ResolutionCache implements IResolutionCache, IConfigurationListener
{
	private int mMaxSize;	
	private long mInterval;
	private long mTouch;
	private Map mResolutionTable = new HashMap(1024);
	private int mGetCount;
	private int mHitCount;
	private AtomicBoolean mPurgeLock=new AtomicBoolean(true);
	private IKernel mKernel;
	private long mLastThreshold;
	private long mLastPurge;
	
	private static ResolutionCache sInstance;
	
	
	public ResolutionCache(IKernel aKernel)
	{	mKernel=aKernel;
		sInstance=this;
		mLastThreshold=System.currentTimeMillis();
	}
	
	public static IResolutionCache getInstance()
	{	return sInstance;
	}
	
	
	public static void removeInstance()
	{	sInstance=null;
	}		
	
	public void setMaxSize(int aSize)
	{	mMaxSize=aSize;
	}
	
	public void putResolution(IRequest aUnresolved, int aUnresolvedScope, IResolution aResolution)
	{
		//safety override if heap gets very tight, release everything
		if (ConcurrentCache.getInstance().isEmergency())
		{	mResolutionTable.clear();
			return;
		}
		
		int requestType=aUnresolved.getVerb();
		if (requestType!=IRequest.VERB_TRANSREPT && mMaxSize>0 && mPurgeLock.get())
		{	long now=mKernel.getTimer().getApproximateTime();
			ResolutionKey k = new ResolutionKey(aUnresolved,aUnresolvedScope,aResolution,now);
			
			if (mPurgeLock.compareAndSet(true, false))
			{	if (mResolutionTable.size()>mMaxSize)
				{	if (now-mLastPurge>mInterval)
					{	mLastPurge=now;
						purge();
					}
				}
				mResolutionTable.put(k, k);
				mPurgeLock.set(true);
			}
		}
	}
	
	public Object[] getResolution(IRequest aUnresolved)
	{	//System.out.println("getResolution "+aUnresolved.toString());
		Object[] result=null;
		
		int requestType=aUnresolved.getVerb();
		if (requestType!=IRequest.VERB_TRANSREPT && mMaxSize>0)
		{
			GetKey getKey=new GetKey(aUnresolved);
			ResolutionKey key = (ResolutionKey)mResolutionTable.get(getKey);
	
			if (key!=null)
			{	long now=mKernel.getTimer().getApproximateTime();
				result = new Object[2];
				result[0]=key.getResolution(aUnresolved,now+mTouch);
				result[1]=aUnresolved.getRequestScope().getDepth()-key.getResolveDepth();
				mHitCount++;
			}
		}
		mGetCount++;
		
		return result;
	}
	
	private void purge()
	{		
		long now=System.currentTimeMillis();
		long threshold=now-(now-mLastThreshold)/4;
		int count=0;
		Set keySet=mResolutionTable.keySet();
		for (Iterator i=keySet.iterator(); i.hasNext(); )
		{	ResolutionKey key = (ResolutionKey)i.next();
			long lru=key.getLRU();
			if (lru<threshold)
			{	i.remove();
				count++;
			}
		}
		mLastThreshold=threshold;
		mGetCount=0;
		mHitCount=0;
				
	}
	
	public String dumpXML() throws IOException, InterruptedException
	{
		StringWriter sw=new StringWriter(1024);
		sw.append("<cache>");
		
		while (!mPurgeLock.compareAndSet(true, false))
		{	Thread.sleep(50);
		}
		
		try
		{
			XMLUtils.writeElement(sw, "gets", Integer.toString(mGetCount),0);
			XMLUtils.writeElement(sw, "hits", Integer.toString(mHitCount),0);
			XMLUtils.writeElement(sw, "size", Integer.toString(mResolutionTable.size()),0);
			
			long now=System.currentTimeMillis();
			for (Iterator i=mResolutionTable.keySet().iterator(); i.hasNext(); )
			{	ResolutionKey key = (ResolutionKey)i.next();
				sw.append("<item>");
				XMLUtils.writeElement(sw, "id", key.getIdentifier(), 0);
				XMLUtils.writeElement(sw, "verb", RequestFactory.verbString(key.getVerb()), 0);
				XMLUtils.writeElement(sw, "lru", Long.toString(now-key.getLRU()), 0);
				XMLUtils.writeElement(sw, "scope", Utils.arrayToString(key.getUnresolvedScope(), 128), 0);
				XMLUtils.writeElement(sw, "transform", Utils.arrayToString(key.getTransform(), 128), 0);
				XMLUtils.writeElement(sw, "endpoint", key.getEndpoint().getClass().getName(), 0);
				sw.append("</item>");
			}
		}
		finally
		{	mPurgeLock.set(true);
		}
		
		sw.append("</cache>");
		
		return sw.toString();
	}

	public void configurationChanged(IConfiguration aConfiguration)
	{	int resolutionCacheSize=aConfiguration.getInt("se.resolutioncache.size", 100);
		mInterval=aConfiguration.getLong("se.resolutioncache.interval", 100);
		mTouch=aConfiguration.getLong("se.resolutioncache.touch", 1000);
		setMaxSize(resolutionCacheSize);
	}
	
	public void clear()
	{	mResolutionTable.clear();
	}
}
